/*
	Copyright (C) 2022 lewis.

	This file is part of ladspa-tool-kit stream extension

	ladspa-tool-kit stream extension is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	ladspa-tool-kit stream extension is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with ladspa-tool-kit stream extension.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/epoll.h>
#include "../include/LTKStream.h"

static int event_poll_fd = -1;
static int num_of_registered_fd = 0;

typedef enum rdwr_operation_e {
	FILE_OPERATION_READ = 0,
	FILE_OPERATION_WRITE = 1
} rdwr_operation_t;

static void ltk_stream_operation(rdwr_operation_t operation,
		int fd, ssize_t *size, char *data, int blocking) {
	/* Be careful on autocast in read and write function */
	/* (tryed size_t and ssize_t, both does not work) */
	int signed_size;

	/* Avoid seg fault */
	if(!size) return;
	if(size[0] < 0) return;
	signed_size = size[0];

	if(blocking) {
		if(operation == FILE_OPERATION_READ) size[0] = 0;

		while(signed_size) {
			ssize_t ret;

			if((ret = (operation == FILE_OPERATION_READ
					? read(fd, data, signed_size)
					: write(fd, data, signed_size))) < 0) {
				size[0] = ret;

				return;
			}

			if(!ret) return;

			data += ret;
			signed_size -= ret;
			if(operation == FILE_OPERATION_READ) size[0] += ret;
		}

		return;
	}

	size[0] = (operation == FILE_OPERATION_READ
		? read(fd, data, signed_size)
		: write(fd, data, signed_size));
}

static void ltk_stream_read(void *obj, va_list arg) {
	ltk_stream_t *stream = (ltk_stream_t *) obj;
	ssize_t *size = va_arg(arg, ssize_t *);
	char *data = va_arg(arg, char *);
	int blocking = va_arg(arg, int);

	/* Perform a read operation */
	ltk_stream_operation(FILE_OPERATION_READ,
		stream->_fd, size, data, blocking
	);
}

static void ltk_stream_write(void *obj, va_list arg) {
	ltk_stream_t *stream = (ltk_stream_t *) obj;
	ssize_t *size = va_arg(arg, ssize_t *);
	char *data = va_arg(arg, char *);
	int blocking = va_arg(arg, int);

	/* Perform a write operation */
	ltk_stream_operation(FILE_OPERATION_WRITE,
		stream->_fd, size, data, blocking
	);
}

static void ltk_stream_print(void *obj, va_list arg) {
	ltk_stream_t *stream = (ltk_stream_t *) obj;
	ssize_t *size = va_arg(arg, ssize_t *);
	char *format = va_arg(arg, char *);
	va_list arguments;
	ssize_t local_size;
	
	va_copy(arguments, arg);
	local_size = vdprintf(stream->_fd, format, arguments);
	if(size) size[0] = local_size;
}

static void ltk_stream_callback(void *obj, va_list arg) {
	ltk_stream_t *stream = (ltk_stream_t *) obj;
	ltk_callback_function_t old_callback_function
		= stream->_callback_function;
	ltk_callback_function_t new_callback_function
		= va_arg(arg, ltk_callback_function_t);
	stream->_callback_data = va_arg(arg, void *);
	uint32_t events = va_arg(arg, uint32_t);

	/* If the request is to add/change a callback */
	if(new_callback_function) {
		/* If no events found, that means that user wants
			to rely on this call to fill this part */
		if(!events) {
			int type = 0;
			char *typename[5] = {
				"block device",
				"symlink",
				"regular file",
				"directory",
				"unknown"
			};

			/* Make sure fd suittable for polling */
			switch(stream->_status.st_mode) {
			default:				/* unknown */
				type++;
			case S_IFDIR:			/* directory */
				type++;
			case S_IFREG:			/* regular file */
				type++;
			case S_IFLNK:			/* symlink */
				type++;
			case S_IFBLK:			/* block device */
				errno = EBADF;

				LTK_LOG_ERROR("Cannot create epoll mechanisms "
					"for %s %p, fd %d : %s",
					typename[type],
					stream,
					stream->_fd,
					strerror(errno)
				);

				return;

			case S_IFCHR:			/* character device */
			case S_IFIFO:			/* FIFO/pipe */
			case S_IFSOCK:			/* socket */
				events |= EPOLLIN;
			}
		}
		/* If events defined, that means that user knows what he does
			therefore, no error check can be performed */

		/* Add/change it */
		stream->_callback_function = new_callback_function;

		/* Subscribe this instance attached to this fd */
		if(!old_callback_function) {
			struct epoll_event event = {events, {stream}};

			if(event_poll_fd < 0) {
				if((event_poll_fd = epoll_create1(0)) < 0) {
					LTK_LOG_ERROR("Cannot create epoll mechanisms : "
						"%s", strerror(errno));

					return;
				}
			}

			if(epoll_ctl(event_poll_fd, EPOLL_CTL_ADD,
					stream->_fd, &event)) {
				LTK_LOG_ERROR("Cannot epoll %p : %s",
					stream, strerror(errno));

				return;
			}

			num_of_registered_fd++;

			LTK_LOG_INFO("Added callback %p on epoll %d "
				"attached to fd %d with flags 0x%X",
				stream, event_poll_fd,
				stream->_fd, events
			);

			LTK_LOG_INFO("Now %d callback(s) registered", num_of_registered_fd);
		}

		LTK_LOG_INFO("Set %p as callback function to %p",
			new_callback_function, stream
		);
	}
	/* If the request is to remove a callback */
	else {
		/* Remove it */
		stream->_callback_function = 0;

		LTK_LOG_INFO("Removed callback function from %p", stream);

		/* And if trigger existed */
		if(old_callback_function) {
			LTK_LOG_INFO("Callback %p was registered, %sremove it",
				stream,
				stream->_free_behaviour
					== LTK_STREAM_FREE_BEHAVIOUR_EPOLL_REMOVE
					? "" : "but don't "
			);
			/* And defined behaviour is to remove it */
			if(stream->_free_behaviour
					== LTK_STREAM_FREE_BEHAVIOUR_EPOLL_REMOVE)
				/* Remove it too */
				if(epoll_ctl(event_poll_fd, EPOLL_CTL_DEL,
						stream->_fd, 0)) {
					/* Unable to remove the trigger,
						keep the callback */
					stream->_callback_function =
						old_callback_function;

					LTK_LOG_ERROR("Unable to remove callback %p from epoll %d",
						stream, event_poll_fd
					);

					return;
				}

			LTK_LOG_INFO("Now %d callback(s) registered",
				num_of_registered_fd - 1
			);

			/* Then treat the global epoll object */
			if(!(--num_of_registered_fd)) {
				LTK_LOG_INFO("No more callback registered, closing epoll %d",
					event_poll_fd
				);

				close(event_poll_fd);

				event_poll_fd = -1;
			}
		}
	}
}

LTKClassTypeDefinition_t *ltk_stream_ctd() {
	static LTKClassTypeDefinition_t *ret = 0;

	if(!ret) {
		LTKFunction_t members[] = {
			ltk_stream_read,
			ltk_stream_write,
			ltk_stream_print,
			ltk_stream_callback,
			0
		};

		ret = LTKAddClass(0, members, sizeof(ltk_stream_t), 0);
	}

	return ret;
}

ltk_stream_t *ltk_stream_new(int fd) {
	ltk_stream_t *ret;
	struct stat buffer;

	if(fstat(fd, &buffer)) {
		LTK_LOG_ERROR("Cannot stat file descriptor %d : %s",
			fd, strerror(errno)
		);

		return 0;
	}

	ret = LTKAdd(ltk_stream_ctd());

	ret->_fd = fd;
	memcpy(&ret->_status, &buffer, sizeof(struct stat));

	return ret;
}

/* Wait for events and call callbacks */
void LTKSync(int timeout) {
	struct epoll_event event;

	LTK_LOG_INFO("Waiting for a stream to trig");

	if(epoll_wait(event_poll_fd, &event, 1, timeout) > 0) {
		ltk_stream_t *stream = event.data.ptr;

		LTK_LOG_INFO("Stream %p trigged", stream);

		if(stream) {
			LTK_LOG_INFO("Launching %p(%p)",
				stream->_callback_function, stream->_callback_data
			);

			stream->_callback_function(stream->_callback_data);

			return;
		}

		LTK_LOG_ERROR("Stream %p does'n contains "
			"any callback function", stream
		);
	}
}

